#include "NodeMCUPins.h"      // NodeMCU Pin Definitions
#include "SPI.h"              // SPI Interface for RFM95
#include "LoRa.h"             // LoRa class & methods
#include "CRC32.h"            // Checksum methods
#include "Wire.h"             // Need this to define second I2C bus
#include "SSD1306Wire.h"      // OLED Display
#include "SketchFonts.h"      // OLED display fonts
#include "SketchImages.h"     // OLED display images
#include "SparkFunBME280.h"   // BME Sensor Library
#include "Ticker.h"           // Ticker class & methods
#include "LangloLoRa.h"       // LoRa configuration parameters
#include "LangloPackets.h"    // Packet structures
#include "nodeB1config.h"     // Node-specific configuration parameters

// NodeMCU to RFM95 (via SPI)
// 3V3                - VCC
// GND                - GND
// NodeMCU_HSCLK (14) - SCK
// NodeMCU_HMISO (12) - MISO
// NodeMCU_HMOSI (13) - MOSI
// NodeMCU_HCS (15)   - NSS
// NodeMCU_D3 (0)     - RST
// NodeMCU_SD3 (10)   - DIO0
// x                  - DIO1
// x                  - DIO2

#define SPI_ID 1              // HSPI — Default SPI configuration

#define SCK  NodeMCU_HSCLK
#define MISO NodeMCU_HMISO
#define MOSI NodeMCU_HMOSI

#define NSS  NodeMCU_HCS
#define RST  NodeMCU_D3
#define DIO0  NodeMCU_SD3

// Packet structures
dataPacket outgoingPacket = {0};
weatherPayload packetPayload = {0};

// NodeMCU to OLED (via I2C)
// GND        - GND
// 3V3        - VCC
// NODEMCU_D1 - SCL
// NODEMCU_D2 - SDA

const int I2C_SCL = NodeMCU_D1;
const int I2C_SDA = NodeMCU_D2;

// OLED Display Settings
const int I2C_Display_Address = 0x3c;
SSD1306Wire display(I2C_Display_Address, I2C_SDA, I2C_SCL);

// BME280 Settings
const int I2C_BME_ADDRESS = 0x77;   // Default SparkFun BME sensor address
BME280 bme;

// Initialise weather variables
uint16_t temperature = 0;
uint16_t pressure = 0;
uint16_t humidity = 0;

const float LSB = 0.02761647; // per manual calibration (when using int in 1/10mm)
uint16_t inputRainfall = 0;
int adc0 = 0;

uint16_t windDirection = 0;
uint8_t windSpeed = 0;

const int reportInterval = 60000;     // Report values every 60000 milliseconds (60 seconds)

// Define the update timer
Ticker dataTicker;

int messageCounter = 0;

void setup() {
  pinMode(NodeMCU_Builtin_LED,OUTPUT);     // Will use LED to signal packet transmission
  digitalWrite(NodeMCU_Builtin_LED, HIGH);  // Turn it off to begin with (although, I don't know why HIGH is off...
  
  Serial.begin(115200);
  while (!Serial);        // If just the the basic function, must connect to a computer
    delay(500);
  Serial.print("[setup] ");
  Serial.print(myName);
  Serial.println(" I2C BME280/LoRa Sensor");
  Serial.println();

// Initialise the OLED display
  display.init();
//  display.flipScreenVertically();
  display.setFont(ArialMT_Plain_10);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(5,5,"BME280/LoRa Sensor");
  display.display();
  
//  Initialise the RFM95 radio
//  SPI.begin(SCK,MISO,MOSI,NSS);
  SPI.begin();                      // Default is SPI1 (HSPI) with SCK (14), MISO (12) & MOSI (13)
//  LoRa.setPins(NSS,RST,DIO0);
  LoRa.setPins(NSS,RST,DIO0);            // No IRQ (DIO0) configured at this time - Not sure why...
  Serial.println("[setup] Activate LoRa Sender");
//  if (!LoRa.begin(BAND, PABOOST)) {
  if (!LoRa.begin(BAND)) {          // No PABOOST in teh ESP8266 impleentation it seems
    Serial.println("[setup] Starting LoRa failed!");
    while (true) {
      yield();                      // This just stops the ESP82665 from getting itself knotted
    }
  }

  Serial.print("[setup] LoRa Frequency: ");
  Serial.println(BAND);
  
  LoRa.setSpreadingFactor(SpreadingFactor);
  LoRa.setSignalBandwidth(SignalBandwidth);
  LoRa.setPreambleLength(PreambleLength);
  LoRa.setCodingRate4(CodingRateDenominator);
  LoRa.setSyncWord(SyncWord);
  LoRa.disableCrc();
  
  Serial.println("[setup] LoRa initialisation complete");
  Serial.println();
  display.drawString(5,20,"LoRa initialisation OK");
  display.display();

// Initialise the BME280 sensor
  Serial.println("[setup] Initialise BME280 sensor");
  bme.setI2CAddress(I2C_BME_ADDRESS);
  if (bme.beginI2C()) {
    Serial.println("[setup] BME sensor initialised");
  } else {
    Serial.println("[setup] Could not find a valid BME280 sensor, check wiring!");
//    while (true){
//      yield();                      // This just stops the ESP82665 from getting itself knotted
//    }
  }
  Serial.println();

  Serial.println("[setup] Set-up Complete");
  Serial.println(""); 

  //  Let everything settle down before continuing
  delay(500);
}

void loop() {
  readSensor();
  Serial.println("[loop] Send Message");
  sendMessage(weatherTypePacket, messageCounter);
  Serial.println("[loop] Update Display");
  updateDisplay(&display, messageCounter);  
  Serial.println("[loop] Increment Message Counter");
  messageCounter++;                 // Increment message count
  Serial.println("[loop] Delay...");
  Serial.println(""); 
  delay(reportInterval);
}

void readSensor() {

  temperature = (int) (10*bme.readTempC());
  pressure = (int) (bme.readFloatPressure() / 91.79F);
  humidity = (int) (bme.readFloatHumidity());

  Serial.println("");
  Serial.print("[readSensor] Temperature: ");
  Serial.println((float) temperature/10, 1);
  Serial.print("[readSensor] Pressure: ");
  Serial.println(pressure);
  Serial.print("[readSensor] Humidity: ");
  Serial.println(humidity);

/*
 * When we've actually got a sensor to read, this variable should be set to the value
 * read from the sensor
 */
//  inputRainfall = (int) adc0 * LSB;
  inputRainfall = 105; // Dummy value (in 1/10 mm) just so taht there's something other than 0 to report

  Serial.print("[readSensor] Rainfall: ");
  Serial.println((float) inputRainfall/10, 1);

/*
 * When we've actually got a sensor to read, these variables should be set to the values
 * read from the sensor(s)
 */
  windDirection = 230;
  windSpeed = 5;

  Serial.print("[readSensor] Wind Direction: ");
  Serial.println(windDirection);
  Serial.print("[readSensor] Wind Speed: ");
  Serial.println(windSpeed);
}

void sendMessage(uint8_t packetType, int counter) {

// If the host is little endian, we need to reverse the byte order before sending

  outgoingPacket.content.destinationMAC = gatewayMAC;
  outgoingPacket.content.sourceMAC = myMAC;
  outgoingPacket.content.sequenceNumber = (uint16_t) counter % 65536;
  outgoingPacket.content.type = packetType;
  outgoingPacket.content.byteCount = (uint8_t) bytesIn(packetType);

  switch (packetType) {
    case powerTypePacket:
// Nothing here yet
      break;
    case tankTypePacket:
// Nothing here yet
      break;
    case pumpTypePacket:
// Nothing here yet
      break;
    case weatherTypePacket:
    {
      weatherPayload payload;
      payload.recorded.temperature = temperature;
      payload.recorded.pressure = pressure;
      payload.recorded.humidity = humidity;

      payload.recorded.rainfall = inputRainfall;

      payload.recorded.windDirection = windDirection;
      payload.recorded.windSpeed = windSpeed;
      
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" weather sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case atmosphereTypePacket:
    {
      atmospherePayload payload;
      payload.recorded.temperature = temperature;
      payload.recorded.pressure = pressure;
      payload.recorded.humidity = humidity;
      Serial.print("[sendMessage] Sending ");
      Serial.print(myName);
      Serial.print(" atmospheric sensor message - Sequence Number ");
      Serial.println(counter % 65536);
      for (int i = 0; i < outgoingPacket.content.byteCount; i++) {
        outgoingPacket.content.payload[i] = payload.payloadByte[i];
      }
      break;
    }
    case rainfallTypePacket:
// Nothing here yet
      break;
    case windTypePacket:
// Nothing here yet
      break;
    case sprinklerTypePacket:
// Nothing here yet
      break;
    default:
      Serial.println("[sendMessage] No payload...");
  }

// Finally, calculate the checksum

  outgoingPacket.content.checksum = generateChecksum(outgoingPacket);

  Serial.println("[sendMessage] Outgoing packet: ");
  digitalWrite(NodeMCU_Builtin_LED, LOW);  // Turn on the LED to signal sending
/*  
// Check that no one else is transmitting first
  while ((bool) LoRa.available()) {
    Serial.println("[sendMessage] Busy...");
// It would take 212 ms to transmit a [max size] 255 byte packet at 9600 bps
    delay(random(200));
  }
*/  
  LoRa.beginPacket();             // start packet
  int totalByteCount = LGO_HeaderSize + outgoingPacket.content.byteCount;
  Serial.print("[sendMessage] Byte Count: ");
  Serial.println( totalByteCount );
  for (int i = 0; i < totalByteCount; i++) {
    LoRa.write(outgoingPacket.packetByte[i]);
    Serial.printf( "%02X", outgoingPacket.packetByte[i]);
    if ((i % 4) == 3) {
      Serial.println("    ");
    } else {
      Serial.print("  ");
    }
  }
  if ((totalByteCount % 4) != 0) Serial.println();

  Serial.println("[sendMessage] Finished Packet");
  LoRa.endPacket();             // Finish packet and send it
  Serial.println("[sendMessage] Packet Sent");
  digitalWrite(NodeMCU_Builtin_LED, HIGH);  // Done
  Serial.println();
}

uint32_t generateChecksum(dataPacket& outgoingPacket) {
  
//  Generate the checksum on the outgoing packet

  CRC32 crc;
  int checksumStart = LGO_HeaderSize - 4;
  int checksumEnd = outgoingPacket.content.byteCount + 4;
  for (int i = checksumStart; i < checksumEnd; i++) {
    crc.update(outgoingPacket.packetByte[i]);
  }
  uint32_t packetChecksum = crc.finalize();
  return packetChecksum;
}

int bytesIn(uint8_t packetType) {
  int result;
  switch (packetType) {
    case powerTypePacket:
      result = powerPayloadBytes;
      break;
    case tankTypePacket:
      result = tankPayloadBytes;
      break;
    case pumpTypePacket:
      result = pumpPayloadBytes;
      break;
    case weatherTypePacket:
      result = weatherPayloadBytes;
      break;
    case atmosphereTypePacket:
      result = atmospherePayloadBytes;
      break;
    case rainfallTypePacket:
      result = rainfallPayloadBytes;
      break;
    case windTypePacket:
      result = windPayloadBytes;
      break;
    case sprinklerTypePacket:
      result = sprinklerPayloadBytes;
      break;
    default:
      result = defaultPayloadBytes;
  }
  return result;
}

void updateDisplay(OLEDDisplay *display, int counter) {
  display->clear();
  display->setTextAlignment(TEXT_ALIGN_CENTER);
  display->setFont(ArialMT_Plain_10);
  display->drawString(64, 0, String(myName) + " Weather");

//  Display the sensor readings
  display->setTextAlignment(TEXT_ALIGN_RIGHT);
  display->drawString(60, 12, "Temperature");
  display->drawString(60, 24, "Pressure");
  display->drawString(60, 36, "Humidity");
  display->setTextAlignment(TEXT_ALIGN_LEFT);
  display->drawString(64, 12, String((float) temperature/10, 1) + " °C");
  display->drawString(64, 24, String(pressure) + " hPa");
  display->drawString(64, 36, String(humidity) + " %");

//  Just display the send message discretely
  display->drawHorizontalLine(0, 50, 128);
  display->drawString(0, 52, "Sending packet");
  display->setTextAlignment(TEXT_ALIGN_RIGHT);
  display->drawString(128, 52, String(counter));

  display->display();
}
